from collections.abc import Mapping
from collections.abc import Sequence
from typing import TypeVar

from .errors import ConfigError
from .errors import MissingKey

T = TypeVar("T")


def parse(value: object, type_: type[T]) -> T:
    if not isinstance(value, type_):
        raise ConfigError(f"Malformed data, expected {type_}")
    return value


def take(data: Mapping[object, object], type_: type[T], key: str) -> T:
    try:
        value = data[key]
    except KeyError as exc:
        raise MissingKey(f"Missing required key: {key!r}") from exc
    return parse(value, type_)


def take_sequence(
    data: Mapping[object, object],
    type_: type[T],
    key: str,
    missing_as_empty: bool = True,
) -> tuple[T, ...]:
    try:
        sequence = data[key]
    except KeyError as exc:
        if missing_as_empty:
            return ()
        raise MissingKey(f"Missing required key: {key!r}") from exc
    if not isinstance(sequence, Sequence):
        raise ConfigError(f"Expected sequence at key {key!r}, got {type(sequence)}")
    return tuple(parse(value, type_) for value in sequence)



from __future__ import annotations

import copy
import hashlib
import re
import sqlite3
import subprocess
import sys
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Sequence
from functools import cache
from pathlib import Path
from sqlite3 import Cursor
from typing import Final
from typing import NewType
from typing import TextIO

import yaml

from . import state_cache
from .errors import ConfigError
from .errors import LocalRepo
from .errors import NoPython
from .parsers import parse
from .parsers import take
from .schema import FQN
from .schema import Checksum
from .schema import Hook
from .schema import Replace
from .schema import Repo
from .schema import Result


@cache
def get_database_cursor() -> Cursor:
    path = (Path.home() / ".cache/pre-commit/db.db").resolve()
    connection = sqlite3.connect(str(path))
    return connection.cursor()


def get_repo_path(fqn: FQN, revision: str) -> Path:
    cursor = get_database_cursor()
    result = cursor.execute(
        """\
        SELECT path FROM repos
        WHERE repo = ? AND ref = ?
        LIMIT 1
        """,
        (fqn, revision),
    )
    (path,) = result.fetchone()
    return Path(path)


ParsedConfig = NewType("ParsedConfig", dict[object, object])


def parse_config(path: Path) -> ParsedConfig:
    return ParsedConfig(parse(yaml.load(path.read_text(), Loader=yaml.CLoader), dict))


def read_configured_repos(config: ParsedConfig) -> Iterator[Repo]:
    repos = config.get("repos", None)
    if not repos or not isinstance(repos, Iterable):
        raise ConfigError(
            "Missing, empty or malformed key `repos` in pre-commit config"
        )
    for repo_data in repos:
        try:
            yield Repo.parse(repo_data)
        except LocalRepo:
            print("Skipping local repo", file=sys.stderr)


def get_unique_hooks(config: ParsedConfig) -> Iterator[tuple[Repo, Hook, FQN]]:
    for repo in read_configured_repos(config):
        seen = set[FQN]()
        for hook in repo.hooks:
            fqn = hook.fully_qualified(repo.repo)
            if fqn in seen:
                continue
            seen.add(fqn)
            yield repo, hook, fqn


def pip_freeze(python_path: Path) -> Iterator[str]:
    command = subprocess.Popen(
        (str(python_path), "-m", "pip", "freeze"),
        stdout=subprocess.PIPE,
    )
    with command as process:
        try:
            return_code = process.wait(timeout=10)
        except:  # noqa: E722 B001
            process.kill()
            raise
        # todo: Handle
        assert return_code == 0
        assert process.stdout is not None

        for line in process.stdout.readlines():
            normalized = line.strip().decode()
            # Filter out reference to the installed hook itself.
            if "@ file" in normalized:
                continue
            yield normalized


def get_python_path(repo_path: Path) -> Path:
    for path in repo_path.glob("py_env*"):
        exec_path = path / "bin/python3"
        if exec_path.exists():
            return exec_path
    raise NoPython


def generate_operations(
    config: ParsedConfig,
    exclude_dependency: str | None,
) -> Iterator[Replace]:
    for repo, hook, fqn in get_unique_hooks(config):
        path = get_repo_path(fqn, repo.rev)

        try:
            python_path = get_python_path(path)
        except NoPython:
            print(f"Missing Python path for {hook.id}", file=sys.stderr)
            continue

        dependencies = tuple(
            dependency
            for dependency in pip_freeze(python_path)
            if (
                exclude_dependency is None
                or not dependency.startswith(f"{exclude_dependency}=")
            )
        )

        if not dependencies:
            continue

        yield Replace(repo=repo, hook=hook, dependencies=dependencies)


version_pattern: Final = re.compile(r"^(?P<name>[\w\-_]+)[<>=]{2}.*$")


def index_dependencies(dependencies: Iterable[str]) -> dict[str, str]:
    return {
        match.group("name"): dependency
        for dependency in dependencies
        if (match := version_pattern.fullmatch(dependency))
    }


def merge_dependencies(
    first_hand: Sequence[str],
    frozen: Sequence[str],
) -> Iterable[str]:
    indexed_first_hand = index_dependencies(first_hand)
    indexed_frozen = index_dependencies(frozen)

    for name, dependency in indexed_first_hand.items():
        if name not in indexed_frozen:
            yield dependency
    yield from frozen


def update_config(config: ParsedConfig, operations: Iterable[Replace]) -> ParsedConfig:
    config = copy.deepcopy(config)
    repos = take(config, list, "repos")

    # Write pinned dependencies.
    for operation in operations:
        try:
            repo = next(
                repo
                for repo in repos
                if operation.repo.repo == repo.get("repo")
                and operation.repo.rev == repo.get("rev")
            )
        except StopIteration:
            raise RuntimeError(
                f"Failed finding repo for append operation {operation.repo.repo=} "
                f"{operation.repo.rev=}"
            )

        try:
            hook = next(
                hook
                for hook in take(repo, list, "hooks")
                if operation.hook.id == take(hook, str, "id")
            )
        except StopIteration:
            raise RuntimeError(
                f"Failed finding hook for append operation {operation.repo.repo=} "
                f"{operation.repo.rev=} {operation.hook.id=}"
            )

        assert isinstance(hook, dict)
        hook["additional_dependencies"] = sorted(
            merge_dependencies(
                first_hand=hook.get("additional_dependencies", ()),
                frozen=operation.dependencies,
            )
        )

    return config


def write_config(config: ParsedConfig, outfile: TextIO) -> None:
    print(
        "# Note! This is an auto-generated file, do not make manual edits here.",
        file=outfile,
    )
    yaml.dump(config, outfile)


def get_checksum(path: Path) -> Checksum:
    return Checksum(hashlib.sha512(path.read_bytes()).hexdigest())


def main(
    outfile_path: Path | None,
    infile_path: Path,
    exclude_dependency: str | None,
    use_cache: bool,
) -> Result:
    outfile_pre_checksum = (
        get_checksum(outfile_path) if outfile_path is not None else None
    )
    state_checksum = f"{get_checksum(infile_path)}-{exclude_dependency=}"

    if (
        use_cache
        and outfile_pre_checksum
        and state_cache.compare(outfile_pre_checksum, state_checksum)
    ):
        print(
            "Found cache match for current state, skipping processing", file=sys.stderr
        )
        return Result.PASSING

    config = parse_config(infile_path)
    operations = tuple(generate_operations(config, exclude_dependency))

    if outfile_path is not None:
        with outfile_path.open("w") as outfile:
            write_config(update_config(config, operations), outfile)

        outfile_post_checksum = get_checksum(outfile_path)

        if outfile_pre_checksum == outfile_post_checksum:
            if use_cache:
                state_cache.write(outfile_post_checksum, state_checksum)
            return Result.PASSING

        return Result.FAILING

    return Result.PASSING if len(operations) == 0 else Result.FAILING



import enum
from collections.abc import Sequence
from dataclasses import dataclass
from typing import NewType
from typing import TypeVar

from .errors import LocalRepo
from .parsers import parse
from .parsers import take
from .parsers import take_sequence

FQN = NewType("FQN", str)
Checksum = NewType("Checksum", str)

# todo: Use typing_extensions.Self when implemented in mypy.
HookSelf = TypeVar("HookSelf", bound="Hook")
RepoSelf = TypeVar("RepoSelf", bound="Repo")


@dataclass(frozen=True, slots=True)
class Hook:
    id: str
    additional_dependencies: Sequence[str]

    @classmethod
    def parse(cls: type[HookSelf], unknown: object) -> HookSelf:
        data = parse(unknown, dict)
        return cls(
            id=take(data, str, "id"),
            additional_dependencies=take_sequence(data, str, "additional_dependencies"),
        )

    def fully_qualified(self, repo: str) -> FQN:
        joined_dependencies = ",".join(sorted(self.additional_dependencies))
        return FQN(
            ":".join((repo, joined_dependencies)) if joined_dependencies else repo
        )


@dataclass(frozen=True, slots=True)
class Repo:
    repo: str
    rev: str
    hooks: Sequence[Hook]

    @classmethod
    def parse(cls: type[RepoSelf], unknown: object) -> RepoSelf:
        data = parse(unknown, dict)
        hooks_data = take(data, list, "hooks")
        repo = take(data, str, "repo")

        if repo == "local":
            raise LocalRepo

        return cls(
            repo=repo,
            rev=take(data, str, "rev"),
            hooks=tuple(Hook.parse(hook_data) for hook_data in hooks_data),
        )


@enum.unique
class Result(enum.Enum):
    PASSING = enum.auto()
    FAILING = enum.auto()
    ERROR = enum.auto()


@dataclass(frozen=True, slots=True, kw_only=True)
class Replace:
    repo: Repo
    hook: Hook
    dependencies: tuple[str, ...]



class UserError(Exception):
    ...


class ConfigError(UserError):
    ...


class MissingKey(ConfigError):
    ...


class NoPython(RuntimeError):
    ...


class LocalRepo(Exception):
    ...


class CacheMiss(RuntimeError):
    ...



import argparse
import sys
from pathlib import Path

from typing_extensions import assert_never

from freeze_transitive.errors import UserError

from . import api
from .schema import Result

parser = argparse.ArgumentParser(
    description="Freeze transitive pre-commit dependencies in Python hooks.",
)
parser.add_argument(
    "--infile",
    type=Path,
    default=Path.cwd() / ".pre-commit-template.yaml",
)
parser.add_argument(
    "--outfile",
    type=Path,
    default=Path.cwd() / ".pre-commit-config.yaml",
)
parser.add_argument(
    "--exclude-dependency",
    type=str,
    default=None,
)
parser.add_argument(
    "--no-cache",
    action="store_true",
)

if __name__ == "__main__":
    args = parser.parse_args()
    outfile_path: Path = args.outfile.resolve()
    infile_path: Path = args.infile.resolve()
    exclude_dependency: str | None = args.exclude_dependency
    use_cache: bool = not args.no_cache

    print(f"Outfile: {str(outfile_path)!r}", file=sys.stderr)
    print(f"Infile: {str(infile_path)!r}", file=sys.stderr)

    try:
        result = api.main(outfile_path, infile_path, exclude_dependency, use_cache)
    except UserError as exc:
        print(f"{exc.__class__.__qualname__}: {exc}", file=sys.stderr)
        result = Result.ERROR

    match result:
        case Result.PASSING:
            pass
        case Result.FAILING:
            exit(1)
        case Result.ERROR:
            exit(2)
        case not_exhaustive:
            assert_never(not_exhaustive)



import sqlite3
from functools import cache
from pathlib import Path
from sqlite3 import Connection

from freeze_transitive import __version__
from freeze_transitive.errors import CacheMiss


def _create_table(connection: Connection) -> None:
    cursor = connection.cursor()
    cursor.execute(
        """\
        CREATE TABLE IF NOT EXISTS state_cache (
            version TEXT NOT NULL,
            key TEXT NOT NULL,
            value TEXT NOT NULL,
            PRIMARY KEY (version, key)
        );
        """
    )


@cache
def _get_database_connection() -> Connection:
    path = (Path.home() / ".cache/pre-commit-freeze-transitive/db.db").resolve()
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
    connection = sqlite3.connect(str(path))
    _create_table(connection)
    return connection


def write(key: str, value: str) -> None:
    connection = _get_database_connection()
    cursor = connection.cursor()
    cursor.execute(
        """\
        INSERT INTO state_cache (version, key, value)
        VALUES (?, ?, ?)
        ON CONFLICT(version, key) DO UPDATE SET
            value=excluded.value;
        """,
        (__version__, key, value),
    )
    connection.commit()


def read(key: str) -> str:
    cursor = _get_database_connection().cursor()
    result = cursor.execute(
        """\
        SELECT value FROM state_cache
        WHERE version = ? AND key = ?
        LIMIT 1
        """,
        (__version__, key),
    )
    row = result.fetchone()
    if row is None:
        raise CacheMiss
    (value,) = row
    assert isinstance(value, str)
    return value


def compare(key: str, value: str) -> bool:
    try:
        return read(key) == value
    except CacheMiss:
        return False
